package se.gawell.kvtstore.core.riak;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import se.gawell.kvtstore.core.DataStoreException;
import se.gawell.kvtstore.core.Deserializer;
import se.gawell.kvtstore.core.EntityNotFoundException;
import se.gawell.kvtstore.core.KeyTimeValueStore;
import se.gawell.kvtstore.core.SecondaryIndexValue;
import se.gawell.kvtstore.core.Serializer;
import se.gawell.kvtstore.core.TimeProvider;

import com.basho.riak.client.http.RiakClient;
import com.basho.riak.client.http.RiakObject;
import com.basho.riak.client.http.request.RequestMeta;
import com.basho.riak.client.http.response.FetchResponse;
import com.basho.riak.client.http.response.IndexResponse;
import com.basho.riak.client.http.response.StoreResponse;

@Component
public class RiakKeyTimeValueStore implements KeyTimeValueStore<String> {

	public static final String CREATION_TIME_INDEX = "creation_time_index_bin";
	private final RiakClient riakClient;
	private final Serializer serializer;
	private final Deserializer deserializer;
	private final TimeProvider timeProvider;
	private final TimeFormatter timeFormatter;

	@Autowired
	public RiakKeyTimeValueStore(RiakClient riakClient, Serializer serializer, Deserializer deserializer,
			TimeProvider timeProvider, TimeFormatter timeFormatter) {
		super();
		this.riakClient = riakClient;
		this.serializer = serializer;
		this.deserializer = deserializer;
		this.timeProvider = timeProvider;
		this.timeFormatter = timeFormatter;
	}

	@Override
	public String put(Object object, SecondaryIndexValue... indexes) {
		StoreResponse storeInRiakResponse = storeInRiak(null, object, indexes);
		return storeInRiakResponse.getObject().getKey();
	}

	@Override
	public void put(String primaryKey, Object object, SecondaryIndexValue... indexes) {
		storeInRiak(primaryKey, object, indexes);
	}

	private StoreResponse storeInRiak(String primaryKey, Object object, SecondaryIndexValue[] indexes) {
		String serializedObject = serializer.serialize(object);
		String key = null;
		if (primaryKey != null) {
			key = primaryKey.toString();
		}
		String bucketNameForClass = getBucketNameForClass(object.getClass());
		
		RiakObject riakObject = new RiakObject(bucketNameForClass, key, serializedObject.getBytes());
		Date now = timeProvider.now();
		riakObject.addIndex(CREATION_TIME_INDEX, timeFormatter.format(now));
		for (SecondaryIndexValue secondaryIndex: indexes) {
			riakObject.addIndex(formatIndexName(secondaryIndex), secondaryIndex.value);
		}
		RequestMeta requestMeta = new RequestMeta();
		requestMeta.setHeader("Content-Type", "application/json");
		StoreResponse storeResponse = riakClient.store(riakObject, requestMeta);
		if (!storeResponse.isSuccess()) {
			throw new DataStoreException("Failure put " + bucketNameForClass + " with key " + primaryKey + ". Message from server was: " + storeResponse.getBodyAsString());
		}
		return storeResponse;
	}

	private String formatIndexName(SecondaryIndexValue secondaryIndex) {
		return secondaryIndex.name.replace(" ", "") + "_bin";
	}

	private String getBucketNameForClass(Class<? extends Object> type) {
		return type.getCanonicalName();
	}

	@Override
	public <type> type getById(String primaryKey, Class<? extends Object> type) {
		String key = primaryKey.toString();
		String bucketNameForClass = getBucketNameForClass(type);
		FetchResponse fetch = riakClient.fetch(bucketNameForClass, key);
		if (fetch.isError()) {
			if (fetch.getStatusCode() == 404) {
				throw new EntityNotFoundException("Could not lookup entity of type " + type + " with primary key " + primaryKey);
			}
			throw new DataStoreException("Failure looking up " + bucketNameForClass + " with key " + primaryKey + ". Message from server was: " + fetch.getBodyAsString());
		}
		String serializedObject = fetch.getObject().getValue();
		return deserializer.deserialize(serializedObject, type);
	}

	@Override
	public <type> List<type> getByTimeInterval(Date startOfIntervalInclusive, Date endOfIntervalExclusive,
			Class<? extends Object> type) {
		String startOfIntervalInclusiveAsString = timeFormatter.format(startOfIntervalInclusive);
		String endOfIntervalExclusiveAsString = timeFormatter.format(endOfIntervalExclusive);
		IndexResponse index = riakClient.index(getBucketNameForClass(type), CREATION_TIME_INDEX, startOfIntervalInclusiveAsString,
				endOfIntervalExclusiveAsString);
		if (index.isError()) {
			throw new DataStoreException("Could not read indexes from riak. " + index.getHttpResponse() + ". Message body: " + createStringFromByteArray(index.getBody()));
		}
		return getListOfObjectsFromListOfIndexes(type, index);
	}

	@SuppressWarnings("unchecked")
	private <type> List<type> getListOfObjectsFromListOfIndexes(Class<? extends Object> type, IndexResponse index) {
		ArrayList<type> list = new ArrayList<type>(index.getKeys().size());
		for (String key: index.getKeys()) {
			list.add((type) getById(key, type));
		}
		return list;
	}

	private String createStringFromByteArray(byte[] bytes) {
		if (bytes != null) {
			return new String(bytes);
		} else {
			return null;
		}
	}


	@Override
	public <type> void delete(String primaryKey, Class<? extends Object> type) {
		try {
		riakClient.delete(getBucketNameForClass(type), primaryKey);
		} catch (RuntimeException e) {
			throw new DataStoreException(e);
		}
	}

	@Override
	public <type> List<type> getByIndex(SecondaryIndexValue index, Class<? extends Object> type) {
		IndexResponse indexResponse = riakClient.index(getBucketNameForClass(type), formatIndexName(index), index.value);
		return getListOfObjectsFromListOfIndexes(type, indexResponse);
	}

}
